#include<stdlib.h>
#include<stdio.h>
#include<string.h>
#include<ctype.h>

#ifdef _WIN32
#define OS_WIN 1
#endif

#ifdef _WIN64
#define OS_WIN 1
#endif

#ifndef OS_WIN
#define OS_LINUX 1
#endif

/***************************/

#include<stdarg.h>
#include<unistd.h>
#include<pthread.h>

#ifdef OS_WIN
#include<errno.h>
#include<io.h>
#include<ctype.h>
#include<direct.h>
#include<windows.h>
#include<setjmp.h>
#include<intrin.h>
#include<sys/timeb.h>

#include"winpthread.h"
#endif


#define __FILENAME__ "compiler.c"


int stricmp(const char *s1, const char *s2) {
/* Yes, Virginia, it had better be unsigned */
unsigned char c1, c2;
while (1) {
    c1 = *s1++;
    c2 = *s2++;
    if (!c1 || !c2)
        break;
    if (c1 == c2)
        continue;
    c1 = tolower(c1);
    c2 = tolower(c2);
    if (c1 != c2)
        break;
}
return (int)c1 - (int)c2;
}



char *TxtReplace(char *in_str, char *in_substr, char *out_substr, int recurse)
{
        char *str_look,*m1,*m2;
        char *out_str;
        int sl,l1,l2,dif,cpt;

        if (in_str==NULL)
                return NULL;

        sl=strlen(in_str);
        l1=strlen(in_substr);
        /* empty string, nothing to do except return empty string */
        if (!sl || !l1)
                return in_str;

        l2=strlen(out_substr);
        dif=l2-l1;

        /* replace string is small or equal in size, we dont realloc */
        if (dif<=0)
        {
                /* we loop while there is a replace to do */
                str_look=strstr(in_str,in_substr);
                while (str_look!=NULL)
                {
                        /* we copy the new string if his len is not null */
                        if (l2)
                                memcpy(str_look,out_substr,l2);
                        /* only if len are different */
                        if (l1!=l2)
                        {
                                /* we move the end of the string byte per byte
                                   because memory locations overlap. This is
                                   faster than memmove */
                                m1=str_look+l1;
                                m2=str_look+l2;
                                while (*m1!=0)
                                {
                                        *m2=*m1;
                                        m1++;m2++;
                                }
                                /* we must copy the EOL */
                                *m2=*m1;
                        }
                        /* look for next replace */
                        if (!recurse)
                                str_look=strstr(str_look+l2,in_substr);
                        else
                                str_look=strstr(in_str,in_substr);
                }
                out_str=in_str;
        }
        else
        {
                /* we need to count each replace */
                cpt=0;
                str_look=strstr(in_str,in_substr);
                while (str_look!=NULL)
                {
                        cpt++;
                        str_look=strstr(str_look+l1,in_substr);
                }
                /* is there anything to do? */
                if (cpt)
                {
                        /* we realloc to a size that will fit all replaces */
                        out_str=realloc(in_str,sl+1+dif*cpt);
                        str_look=strstr(out_str,in_substr);
                        while (str_look!=NULL && cpt)
                        {
                                /* as the replace string is bigger we
                                   have to move memory first from the end */
                                m1=out_str+sl;
                                m2=m1+dif;
                                sl+=dif;
                                while (m1!=str_look+l1-dif)
                                {
                                        *m2=*m1;
                                        m1--;m2--;
                                }
                                /* then we copy the replace string (can't be NULL in this case) */
                                memcpy(str_look,out_substr,l2);

                                /* look for next replace */
                                if (!recurse)
                                        str_look=strstr(str_look+l2,in_substr);
                                else
                                        str_look=strstr(in_str,in_substr);

                                /* to prevent from naughty overlap */
                                cpt--;
                        }
                        if (str_look!=NULL)
                        {
                                printf("INTERNAL ERROR - overlapping replace string (%s/%s), you can't use this one!\n",in_substr,out_substr);
                                exit(-1);
                        }
                }
                else
                        out_str=in_str;
        }
        return out_str;
}



void FieldArrayAddDynamicValueConcat(char ***zearray, int *nbfields, int *maxfields, char *zevalue)
{
        if ((*zearray)==NULL) {
                *nbfields=1;
                *maxfields=10;
                (*zearray)=realloc(NULL,(*maxfields)*sizeof(char *));
        } else {
                *nbfields=(*nbfields)+1;
                if (*nbfields>=*maxfields) {
                        *maxfields=(*maxfields)*2;
                        (*zearray)=realloc((*zearray),(*maxfields)*sizeof(char *));
                }
        }
        /* using direct calls because it is more interresting to know which is the caller */
        (*zearray)[(*nbfields)-1]=strdup(zevalue);
        (*zearray)[(*nbfields)]=NULL;
}


struct s_parameter {
	char *scriptfile;
};

void IntArrayAddDynamicValueConcat(int **zearray, int *nbval, int *maxval, int zevalue)
{
        if ((*zearray)==NULL) {
                *nbval=1;
                *maxval=4;
                (*zearray)=malloc(sizeof(int)*(*maxval));
        } else {
                *nbval=*nbval+1;
                if (*nbval>=*maxval) {
                        *maxval=(*maxval)*2;
                        (*zearray)=realloc((*zearray),sizeof(int)*(*maxval));
                }
        }
        (*zearray)[(*nbval)-1]=zevalue;
}
char **TxtSplitWithChar(char *in_str, char split_char, int *nbw)
{
        #undef FUNC
        #define FUNC "TxtSplitWithChar"

        char **tab=NULL;
        char *match_str;
        int redo=1;
        int idx,idmax;

        match_str=in_str;

        while (redo && *in_str) {
                while (*match_str && *match_str!=split_char) match_str++;
                redo=*match_str;
                *match_str=0;
                FieldArrayAddDynamicValueConcat(&tab,&idx,&idmax,in_str);
                in_str=++match_str;
        }

        *nbw=idx;
        return tab;
}

char *ConcatWords(char **wordlist) {
        char *strc=NULL;
        int i=0;

        strc=strdup(wordlist[i]);
        while (wordlist[i+1]) {
                strc=realloc(strc,strlen(strc)+1+strlen(wordlist[i+1]+1));
                strcat(strc," ");
                strcat(strc,wordlist[i+1]);
                i++;
        }
        return strc;
}

void FreeWords(char **myw) {
        int i=0;
        while (myw[i]) {
                free(myw[i]);
                i++;
        }
        free(myw);
}


/************************ multi-threaded execution *************************************************/

void ObjectArrayAddDynamicValueConcat(void **zearray, int *nbfields, int *maxfields, void *zeobject, int object_size)
{
	#undef FUNC
	#define FUNC "ObjectArrayAddDynamicValueConcat"

	char *dst;

	if ((*zearray)==NULL) {
		*nbfields=1;
		*maxfields=8;
		(*zearray)=malloc((*maxfields)*object_size);
	} else {
		*nbfields=(*nbfields)+1;
		if (*nbfields>=*maxfields) {
			*maxfields=(*maxfields)*2;
			(*zearray)=realloc((*zearray),(*maxfields)*object_size);
		}
	}
	/* using direct calls because it is more interresting to know which is the caller */
	dst=((char *)(*zearray))+((*nbfields)-1)*object_size;
	/* control of memory for overflow */
	memcpy(dst,zeobject,object_size);
}


int _internal_CPUGetCoreNumber()
{
        #undef FUNC
        #define FUNC "_internal_CPUGetCoreNumber"

        static int nb_cores=0;
#ifdef OS_WIN
		SYSTEM_INFO sysinfo;
		if (!nb_cores) {
			GetSystemInfo( &sysinfo );
			nb_cores=sysinfo.dwNumberOfProcessors;
			if (nb_cores<=0)
					nb_cores=1;
		}
#else
        if (!nb_cores) {
                nb_cores=sysconf(_SC_NPROCESSORS_ONLN );
                if (nb_cores<=0)
                        nb_cores=1;
        }
#endif
        return nb_cores;
}
#define CPU_NB_CORES _internal_CPUGetCoreNumber()

void diff_printf(char **output, int *lenoutput,...)
{
	#undef FUNC
	#define FUNC "diff_printf"
	
	char tmp[2048];
	int curlen;
	char *format;
	va_list argptr;

	va_start(argptr,lenoutput);
	format=va_arg(argptr,char *);
	vsprintf(tmp,format,argptr);
	curlen=strlen(tmp); // windows compliant...
	va_end(argptr);
	
	*output=realloc(*output,(*lenoutput)+curlen+1);
//printf("input=[%s] tmp=[%s] curlen=%d\n",*output,tmp,curlen);
	memcpy((*output)+*lenoutput,tmp,curlen);
//printf("output=[%s]\n",*output);
	*lenoutput=(*lenoutput)+curlen;
}


struct s_compilation_action {
	unsigned char *sprite;
	unsigned char *mask;
	int width,height;
	char *label;
};

struct s_compilation_thread {
	pthread_t thread;
	struct s_compilation_action *action;
	int nbaction,maxaction;
	char *output;
	int lenoutput;
};

struct s_compilation_thread *SplitForThreads(int nbwork, struct s_compilation_action *ca, int *nb_cores)
{
	#undef FUNC
	#define FUNC "SplitForThread"

	struct s_compilation_thread *compilation_threads;
	int i,j,curidx,step;

	/* sometimes we do not want to split the work into the exact number of cores */
	if (!*nb_cores) {
		*nb_cores=_internal_CPUGetCoreNumber();
	}
	compilation_threads=calloc(1,sizeof(struct s_compilation_thread)*(*nb_cores));

	/* dispatch workload */
	step=nbwork/(*nb_cores);
	while (step<2 && *nb_cores>2) {
		*nb_cores=*nb_cores>>1;
		step=nbwork/(*nb_cores);
	}
	if (step<1) step=1;

	while (step*(*nb_cores)<nbwork) step++;

	fprintf(stderr,"split %d tasks with %d core%s and average step of %d\n",nbwork,*nb_cores,*nb_cores>1?"s":"",step);

	/* split image into strips */
	for (curidx=i=0;i<*nb_cores;i++) {
		for (j=0;j<step && curidx<nbwork;j++) {
			ObjectArrayAddDynamicValueConcat((void **)&compilation_threads[i].action,&compilation_threads[i].nbaction,&compilation_threads[i].maxaction,&ca[curidx++],sizeof(struct s_compilation_action));
		}
	}
	// we might be in the case where there is task left...
	while (curidx<nbwork) {
		printf("%d task missing, dispatching again\n",nbwork-curidx);
		for (i=0;i<*nb_cores;i++) {
			if (curidx<nbwork) {
				ObjectArrayAddDynamicValueConcat((void **)&compilation_threads[i].action,&compilation_threads[i].nbaction,&compilation_threads[i].maxaction,&ca[curidx++],sizeof(struct s_compilation_action));
			}
		}
	}
	for (i=0;i<*nb_cores;i++) {
		printf("core %3d : %5d task%s",i,compilation_threads[i].nbaction,compilation_threads[i].nbaction>1?"s":"");
		for (j=0;j<compilation_threads[i].nbaction;j++) {
			if (strcmp("herbal_base",compilation_threads[i].action[j].label)==0) {
				printf(" => herbal in %d position",j);
			}
		}
		printf("\n");
	}

	return compilation_threads;
}

void CompileEntity(char **output, int *lenoutput, struct s_compilation_action *);

void *MakeDiffThread(void *param)
{
	#undef FUNC
	#define FUNC "MakeDiffThread"

	int i,j;
	struct s_compilation_thread *ct;
	ct=(struct s_compilation_thread *)param;
	for (i=0;i<ct->nbaction;i++) {
		//if (strcmp(ct->action[i].label,"megastore_left")==0) 
		CompileEntity(&ct->output, &ct->lenoutput,&ct->action[i]);
	}
	pthread_exit(NULL);
	return NULL;
}

void ExecuteThreads(int nb_cores, struct s_compilation_thread *compilation_threads, void *(*fct)(void *))
{
	#undef FUNC
	#define FUNC "ExecuteThreads"

	pthread_attr_t attr;
	void *status;
	int i,rc;
	/* launch threads */
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_JOINABLE);
	pthread_attr_setstacksize(&attr,655360);
 	for (i=0;i<nb_cores;i++) {
		if ((rc=pthread_create(&compilation_threads[i].thread,&attr,fct,(void *)&compilation_threads[i]))) {
			fprintf(stderr,"Cannot create thread!");
			exit(-5);
		}
	}
 	for (i=0;i<nb_cores;i++) {
		if ((rc=pthread_join(compilation_threads[i].thread,&status))) {
			fprintf(stderr,"thread error!");
			exit(-5);
		}
	}
}





void __FileReadBinary(char *filename, char *data, int len, int curline)
{
	FILE *f;

	f=fopen(filename,"rb");
	if (!f) {
		printf("(%d) file [%s] not found or cannot be opened!\n",curline,filename);
		exit(-1);
	}
	if (fread(data,1,len,f)!=len) {
		printf("(%d) file [%s] cannot read %d byte(s)!\n",curline,filename,len);
		exit(-1);
	}
	fclose(f);
}
#define FileReadBinary(fichier,data,len) __FileReadBinary(fichier,data,len,__LINE__)

#define NBREG 4

char *GetValStr(char *txtbuffer, int *current_reg,int v) {
        #undef FUNC
        #define FUNC "GetValStr"

	static char *txtreg[NBREG]={"b","c","d","e"};
	int i;
//fprintf(stderr,"#%02X,",v);
	for (i=0;i<NBREG;i++) {
		if (v==current_reg[i]) {
			sprintf(txtbuffer,"%s",txtreg[i]);
			return txtbuffer;
		}
	}
	sprintf(txtbuffer,"#%02X",v);
	return txtbuffer;
}


struct s_sprval {
        int val;
        int cpt;
        int prv;
};
struct s_regupdate {
        int reg;
        int val;
        int oldval;
};
int compare_sprval(const void *a, const void *b)
{
        struct s_sprval *sa,*sb;
        sa=(struct s_sprval *)a;
        sb=(struct s_sprval *)b;
        /* du plus grand au plus petit */
        return sb->cpt-sa->cpt;
}


// optimizer v2
struct s_cell {
        struct s_cell *prev,*next;
        int val;
        int cpt;
};

struct s_flux {
        int val;
        int cpt;
};
int compare_fluxcount(const void *a, const void *b)
{
        struct s_flux *sa,*sb;
        sa=(struct s_flux *)a;
        sb=(struct s_flux *)b;
        /* du plus grand au plus petit */
        return sb->cpt-sa->cpt;
}

int *ComputeDiffStats(char **output, int *lenoutput,unsigned char *flux, int start, int size, int *cval) {
        struct s_flux fcount[256];
        struct s_flux fbefor[256];
        int i,j,k,l,mreg;
        char *reg[]={"b","c","d","e"};

int dibiji=0;
//if (strstr(*output,"construction03_left")) dibiji=1;

        if (!start) {
                for (i=0;i<NBREG;i++) {
                        cval[i]=-1;
                }
        }
        // count bytes
        for (i=0;i<256;i++) {
                fcount[i].val=i;
                fcount[i].cpt=0;
                fbefor[i].val=i;
                fbefor[i].cpt=0;
        }
        for (i=start;i<size;i++) {
                fcount[flux[i]].cpt++;
        }
        qsort(fcount,256,sizeof(struct s_flux),compare_fluxcount);

        // first init avec les premières données à dépasser le seuil
        if (!start) {
                for (i=mreg=0;i<size;i++) {
                        fbefor[flux[i]].cpt++;
                        if (fbefor[flux[i]].cpt>4) {
                                // init!
                                cval[mreg++]=flux[i];
                                fbefor[flux[i]].cpt=-size-1; // annuler les prochains tests
                                if (mreg==NBREG) break;
                        }
                }
                // reinit for 2nd pass
                for (i=0;i<256;i++) {
                        if (fbefor[i].cpt>=0) fbefor[i].cpt=0; // mais pas les valeurs déjà validées
                        else fbefor[i].cpt-=size; // on en remet un petit coup!
                }
                for (i=0;i<size;i++) {
                        fbefor[flux[i]].cpt++;
                        if (fbefor[flux[i]].cpt>2) {
                                // init!
                                cval[mreg++]=flux[i];
                                fbefor[flux[i]].cpt=-size; // annuler les prochains tests
                                if (mreg==NBREG) break;
                        }
                }
                i=0;
                while (i+1<NBREG) {
                        if (cval[i]>=0 && cval[i+1]>=0) {
                                diff_printf(output,lenoutput,"ld %s%s,#%02X%02X\n",reg[i],reg[i+1],cval[i],cval[i+1]);
                        } else if (cval[i]>=0) {
                                diff_printf(output,lenoutput,"ld %s,#%02X\n",reg[i],cval[i]);
                        }
                        i+=2;
                }
                if (i<NBREG) {
                        if (cval[i]>=0) diff_printf(output,lenoutput,"ld %s,#%02X\n",reg[i],cval[i]);
                }
        } else {
                int inplace,imax,rmax,tresh,ireplace;

                // tester si la valeur courante est en cache
                for (i=inplace=0;i<NBREG;i++) {
                        if (flux[start]==cval[i]) {
                                inplace=1;
                                break;
                        }
                }

                /* on va, pour toutes les valeurs >4 qui ne sont pas déjà en cache,
                 * regarder si on peut se substituer à un des caches existants
                 */

                // si elle n'est pas en cache alors on va voir si on peut changer un registre, celui qui est le plus loin de la position courante
                if (!inplace) {
                        // retrouver la valeur courante dans le tri
                        for (i=0;i<256;i++) {
                                if (fcount[i].val==flux[start]) break;
                        }
                        ireplace=i;

                        // en fonction de la quantité on change le seuil (car on pourrait avoir besoin de rebasculer sur une autre valeur et ça a un coût)
                        switch (fcount[NBREG-1].cpt) {
                                case 0:tresh=2;break;
                                case 1:tresh=3;break;
                                default:tresh=4;break;
                        }

                        // est-ce que la valeur dont on a besoin a intérêt à être remplacée? est-ce qu'on dépasse notre seuil?
                        if (fcount[ireplace].cpt>tresh) {
                                if (dibiji) printf("valeur #%02X on dépasse notre seuil %d>%d\n",fcount[ireplace].val,fcount[ireplace].cpt,tresh);

                                // on cherche parmi les valeurs en cache celle qui est le plus loin possible de la courante pour augmenter nos chances de remplacement  
                                for (j=imax=0;j<NBREG;j++) {
                                        for (i=start;i<size;i++) {
                                                if (flux[i]!=cval[j]) {
						       	if (i>imax) {
	                                                        imax=i;
        	                                                rmax=j;
							}
                                                } else break;
                                        }
                                }
                                if (dibiji) printf(" rmax=%d candidate register=%s imax=%d\n",rmax,reg[rmax],imax);

                                // ensuite on va analyser la distribution dans le flux au plus proche pour voir si c'est possible de remplacer
                                for (i=start;i<size;i++) {
                                        fbefor[flux[i]].cpt++;

                                        // tant que la valeur à remplacer ne dépasse pas le seuil
                                        if (fbefor[cval[rmax]].cpt<tresh+1) {
                                                // si la valeur en cache dépasse ou égalise la plus balaise avant le seuil, alors on n'en tient pas compte
                                                if (fbefor[cval[rmax]].cpt>=fbefor[flux[start]].cpt) {
                                                        if (dibiji) printf(" valeur en cache > nouvelle candidate avant le seuil\n");
                                                        break;
                                                }
                                        } else {
						// on a un premier feu vert pour remplacer la valeur par la nouvelle MAIS
						// est-ce qu'il n'existerait pas d'autres valeurs +intéressantes sur l'intervale?
						// qui ne soient pas en cache...
						for (l=0;l<256;l++) {
							if (l==flux[start]) continue;
							if (fbefor[l].cpt>fbefor[flux[start]].cpt) {
								// on a un meilleur candidat sur l'intervale!
								for (j=0;j<NBREG;j++) {
									if (cval[j]==l) break;
								}
								// et il n'est pas en cache!
								if (j==NBREG) {
									if (dibiji) printf("pas de remplacement car on a mieux sur la période!\n");
									break;
								}
							}
						}
						if (l==256) {
							if (dibiji) printf("stat update because #%02X tresh at %d\n",fbefor[cval[rmax]].val,i);
        	                                        // la nouvelle valeur dépasse le seuil d'occurences
                	                                diff_printf(output,lenoutput,"\n/* stat update */ ld %s,#%02X ; idx=%d\n",reg[rmax],flux[start],start);
                        	                        cval[rmax]=flux[start];
						}
						// else break dans tous les cas!
                                                break;
                                        }
                                }
				if (i==size) {
	                                if (dibiji) printf("analyse de flux n'a pas rencontré assez de valeurs à remplacer,c'est qu'on peut écraser! (ou pas!)\n");
					for (l=0;l<256;l++) {
						if (l==flux[start]) continue;
						if (fbefor[l].cpt>fbefor[flux[start]].cpt) {
							// on a un meilleur candidat sur l'intervale!
							for (j=0;j<NBREG;j++) {
								if (cval[j]==l) break;
							}
							// et il n'est pas en cache!
							if (j==NBREG) {
								if (dibiji) printf("pas de remplacement car on a mieux sur la période!\n");
								break;
							}
						}
					}
					if (l==256) {
						if (dibiji) printf("validé sur la période on remplace\n");
        		                        diff_printf(output,lenoutput,"\n/* stat update */ ld %s,#%02X ; idx=%d\n",reg[rmax],flux[start],start);
	                	                cval[rmax]=flux[start];
					}
				}
                        } else {
                                if (dibiji) printf("valeur courante pas en cache et pas intéressante (x%d)\n",fcount[ireplace].cpt);
                        }
                }
        }

        return cval;
}
int *_old_ComputeDiffStats(char **output, int *lenoutput,unsigned char *flux, int start, int size, int *current_reg)
{
        /******************************************************
             greatly inspired by hardware sprite compiler 
        ******************************************************/
	struct s_regupdate regupdate[NBREG];
	struct s_sprval switchspr;
	struct s_sprval diff[256];
	struct s_sprval shortdiff[256];
	int old_reg[NBREG],new_reg[NBREG];
	int ireg=0;
	int i,j,k,l;
	int kk,change=0;
	int treshold;

//#define STAT_INFO 1

        /******************************************************
                   i n i t i a l i s e    s t a t s
        ******************************************************/
        /* backup reg */
        for (i=0;i<NBREG;i++) old_reg[i]=current_reg[i];

        /* stat struct */
        for (i=0;i<256;i++) {
                diff[i].val=i;
                diff[i].cpt=0;
                diff[i].prv=0;
        }

        /* where are the previous values in the stats? */
        for (i=0;i<NBREG;i++) {
                for (j=0;j<256;j++) {
                        if (diff[j].val==old_reg[i]) {
                                diff[j].prv=1;
                                break;
                        }
                }
        }

        /******************************************************
               d o i n ' n   s o r t i n '      s t a t s
        ******************************************************/
	/* le comptage proprement dit (peut évoluer en différences de flux à CET ENDROIT mais peu d'usage hors démo) */
        for (i=start;i<size;i++) {
                diff[flux[i]].cpt++;
        }
	qsort(diff,256,sizeof(struct s_sprval),compare_sprval);

#if STAT_INFO
        fprintf(stderr,"**** before optim - start=%d\n",start);
        for (i=0;i<NBREG*2;i++) {
                if (!diff[i].cpt) break;
                fprintf(stderr,"diff[%d] val=%d cpt=%d\n",i,diff[i].val,diff[i].cpt);
        }
#endif

        /******************************************************
              r e o p t i m i s e    s t a t s
        ******************************************************/
        /* especially at the beginning of the sprite we need to look forward for a better optimisation*/

	if (diff[NBREG-1].cpt>3) {
#if STAT_INFO
		fprintf(stderr,"-> try optimisation\n");
#endif
                /* can we change a secondary color before the use of a major color? */
                for (j=0;j<NBREG;j++) {
			/* dans combien d'octets se situe l'occurence? */
                        for (i=start;i<size;i++) {
                                if (flux[i]==diff[j].val) {
                                        break;
                                }
                        }
                        /* do stats for changes until the marker i */
                        for (l=0;l<256;l++) {
                                shortdiff[l].val=diff[l].val;
                                shortdiff[l].cpt=0;
                                shortdiff[l].prv=0; // toujours ZERO a ce stade => diff[l].prv;
                        }
                        /* stats for secondary colors only */
                        for (l=start;l<i;l++) {
                                for (k=5;k<256;k++) {
                                        if (flux[l]==shortdiff[k].val) shortdiff[k].cpt++;
                                }
                        }
                        qsort(shortdiff,256,sizeof(struct s_sprval),compare_sprval);

#if STAT_INFO
                        printf("stats de %d a %d pour la couleur %d\n",start,i,diff[j].val);
                        for (l=0;l<NBREG;l++) {
                                if (!shortdiff[l].cpt) break;
                                printf("shortdiff[%d] val=%d cpt=%d\n",l,shortdiff[l].val,shortdiff[l].cpt);
                        }
#endif

			/* v1.3 bugfixed => ONLY WITH A REGISTER USAGE */
			//if (((shortdiff[0].val+1)&255)==diff[j].val || ((shortdiff[0].val-1)&255)==diff[j].val) treshold=2; else treshold=4;
			treshold=4;

			/* on conditionne le takeover à la valeur qui arrive de suite! */
			if (shortdiff[0].cpt>=treshold && (shortdiff[0].val==flux[start] || !start)) {
				/* takeover on bigger count */
				for (k=0;diff[k].val!=shortdiff[0].val;k++);
				switchspr=diff[j];
				diff[j]=diff[k];
				diff[k]=switchspr;
			}
#if STAT_INFO
		diff_printf(output,lenoutput,"; LEGACY surcharge possible de val=%d cpt=%d\n",shortdiff[0].val,shortdiff[0].cpt);
		diff_printf(output,lenoutput,"; couleur à surcharger #%02X par #%02X\n",diff[j].val,diff[k].val);
#endif

                }
        }

#if STAT_INFO
fprintf(stderr,"**** after optim\n");
        for (i=0;i<NBREG*2;i++) {
                if (!diff[i].cpt) break;
                fprintf(stderr,"diff[%d] val=%d cpt=%d prv=%d\n",i,diff[i].val,diff[i].cpt,diff[i].prv);
        }
#endif
        /******************************************************
              i n i t i a l i s e r
        ******************************************************/
        if (!start) {
                for (i=0;i<NBREG;i++) {
                        //printf("diff[%d].cpt=%d val=%d\n",i,diff[i].cpt,diff[i].val);
                        if (diff[i].cpt<3) {
                                current_reg[i]=diff[i].val=-1;
                        } else {
                                current_reg[i]=diff[i].val;
                        }
                }
#if 0
		// no A register use
		for (i=1;i<5;i++) {
			if (!diff[i].val) {
				/* v1.3 */
                        	diff_printf(output,lenoutput,"; optimise zero for A register at init\n");
				current_reg[i]=diff[i].val=diff[0].val;
				current_reg[0]=diff[0].val=0;
			}
		}
#endif
                if (diff[0].val>=0 && diff[1].val>=0) diff_printf(output,lenoutput,"ld bc,#%X\n",diff[0].val*256+diff[1].val); else
                {
                        if (diff[0].val>=0) diff_printf(output,lenoutput,"ld b,%d\n",diff[0].val); else
                        if (diff[1].val>=0) diff_printf(output,lenoutput,"ld c,%d\n",diff[1].val);
                }
                if (diff[2].val>=0 && diff[3].val>=0) diff_printf(output,lenoutput,"ld de,#%X\n",diff[2].val*256+diff[3].val); else
                {
                        if (diff[2].val>=0) diff_printf(output,lenoutput,"ld d,%d\n",diff[2].val);
                        if (diff[3].val>=0) diff_printf(output,lenoutput,"ld e,%d\n",diff[3].val);
                }

		/* no override after init */
                return current_reg;
        }

        /******************************************************
              b u f f e r e d    u p d a t e r
        ******************************************************/
        for (i=0;i<NBREG;i++) {
                /* looking for the best slot */
                if (!diff[i].prv) {
			/* optim v1.4 */
			if (diff[i].val==flux[start]) {
				/* override a prv with stat-3 */
				for (j=255;j>=0;j--) {
					/* look from the end of stats v1.3 */
					if (diff[j].prv && diff[i].cpt-3>diff[j].cpt) {
						/* need to know which register will be replaced */
						for (k=0;k<NBREG;k++) {
							if (current_reg[k]==diff[j].val) break;
						}
						if (k==NBREG) {
							fprintf(stderr,"bug dans la reaffectation des registres\n");
							exit(-1);
						}

						/* register change is buffered */
						regupdate[ireg].reg=k;
						regupdate[ireg].val=diff[i].val;
						regupdate[ireg].oldval=diff[j].val;
						ireg++;
						/* confirm register change */
						current_reg[k]=diff[i].val;
						diff[i].prv=1;
						break;
					}
				}
			}
                }
        }
        /******************************************************
              u p d a t e
        ******************************************************/
        if (ireg) {
                int ib=-1,ic=-1,id=-1,ie=-1;

                diff_printf(output,lenoutput,"\n/* stats update */ ");
#if STAT_INFO
printf("update registers\n");
#endif

                /* get B,C,D,E indexes */
                for (i=0;i<ireg;i++) {
                        switch (regupdate[i].reg) {
                                case 0:ib=i;break;
                                case 1:ic=i;break;
                                case 2:id=i;break;
                                case 3:ie=i;break;
                                default:diff_printf(output,lenoutput,"register error in update\n");exit(-1);
                        }
                }
                /* try to pack BC */
                if (ib>=0 && ic>=0) {
                        int packed=1;
                        if ((regupdate[ib].val==(regupdate[ib].oldval-1)&0xFF) || (regupdate[ib].val==(regupdate[ib].oldval+1)&0xFF)) {
                                if ((regupdate[ic].val==(regupdate[ic].oldval-1)&0xFF) || (regupdate[ic].val==(regupdate[ic].oldval+1)&0xFF)) {
                                        /* do not pack BC for thoses values 1.2 */
                                        if (regupdate[ib].val==(regupdate[ib].oldval-1)&0xFF) diff_printf(output,lenoutput,"dec b :"); else diff_printf(output,lenoutput,"inc b :");
                                        if (regupdate[ic].val==(regupdate[ic].oldval-1)&0xFF) diff_printf(output,lenoutput,"dec c\n"); else diff_printf(output,lenoutput,"inc c\n");
                                        packed=0;
                                }
                        }
                        if (packed) diff_printf(output,lenoutput,"ld bc,#%X\n",regupdate[ib].val*256+regupdate[ic].val);
                } else {
                        if (ib>=0) {
                                if (regupdate[ib].val==(regupdate[ib].oldval-1)&0xFF) diff_printf(output,lenoutput,"dec b\n"); else
                                if (regupdate[ib].val==(regupdate[ib].oldval+1)&0xFF) diff_printf(output,lenoutput,"inc b\n"); else diff_printf(output,lenoutput,"ld b,%d\n",regupdate[ib].val);
                        }
                        if (ic>=0) {
                                if (regupdate[ic].val==(regupdate[ic].oldval-1)&0xFF) diff_printf(output,lenoutput,"dec c\n"); else
                                if (regupdate[ic].val==(regupdate[ic].oldval+1)&0xFF) diff_printf(output,lenoutput,"inc c\n"); else diff_printf(output,lenoutput,"ld c,%d\n",regupdate[ic].val);
                        }
                }
                /* try to pack DE */
                if (id>=0 && ie>=0) {
                        /* (fixed v1.2) */
                        int packed=1;
                        if ((regupdate[id].val==(regupdate[id].oldval-1)&0xFF) || (regupdate[id].val==(regupdate[id].oldval+1)&0xFF)) {
                                if ((regupdate[ie].val==(regupdate[ie].oldval-1)&0xFF) || (regupdate[ie].val==(regupdate[ie].oldval+1)&0xFF)) {
                                        /* do not pack DE for thoses values 1.2 */
                                        if (regupdate[id].val==(regupdate[id].oldval-1)&0xFF) diff_printf(output,lenoutput,"dec d :"); else diff_printf(output,lenoutput,"inc d :");
                                        if (regupdate[ie].val==(regupdate[ie].oldval-1)&0xFF) diff_printf(output,lenoutput,"dec e\n"); else diff_printf(output,lenoutput,"inc e\n");
                                        packed=0;
                                }
                        }
                        if (packed) diff_printf(output,lenoutput,"ld de,#%X\n",regupdate[id].val*256+regupdate[ie].val);
                } else {
                        if (id>=0) {
                                if (regupdate[id].val==(regupdate[id].oldval-1)&0xFF) diff_printf(output,lenoutput,"dec d\n"); else
                                if (regupdate[id].val==(regupdate[id].oldval+1)&0xFF) diff_printf(output,lenoutput,"inc d\n"); else diff_printf(output,lenoutput,"ld d,%d\n",regupdate[id].val);
                        }
                        if (ie>=0) {
                                if (regupdate[ie].val==(regupdate[ie].oldval-1)&0xFF) diff_printf(output,lenoutput,"dec e\n"); else
                                if (regupdate[ie].val==(regupdate[ie].oldval+1)&0xFF) diff_printf(output,lenoutput,"inc e\n"); else diff_printf(output,lenoutput,"ld e,%d\n",regupdate[ie].val);
                        }
                }
        }
        return current_reg;
}


void CompileEntity(char **output, int *lenoutput, struct s_compilation_action *action)
{
        #undef FUNC
        #define FUNC "CompileEntity"

	int i,j,first=1,x,y,dx;
	int previous_l_value=500;
	char txtbuffer[32];
	int current_reg[NBREG]={-1,-1,-1,-1};
	int width,height,iflux=0,mflux,pix;
	unsigned char *sprite,*flux,*mask;
	int linecpt,sens;
	int repzone[16];
	int repcount[16];
	int skiptrigger=0;

	int debuginfo=0;
	int skiplines,skiplines_ref;

	#define GAUCHE 0xAA
        #define DROITE 0x55

//	if (strstr(action->label,"construction03_left")) debuginfo=1;

#define CHECKFLUX(a) if (a!=flux[iflux]

	sprite=action->sprite;
	mask=action->mask;
	width=action->width;
	height=action->height;

	flux=malloc(width*height*4); // grand max!
	

//fprintf(stderr,"%s processing (len=%d)\n",action->label,*lenoutput);

	diff_printf(output,lenoutput,";************************************\n");
	diff_printf(output,lenoutput,"  %s\n",action->label);
	diff_printf(output,lenoutput,";************************************\n");

	linecpt=0;

	// skip lower empty lines
	for (y=height-1;y>0;y--) {
		for (x=0;x<width;x++) {
			if (mask[(x+y*width)*2] || mask[(x+y*width)*2+1]) {
				break;
			}
		}
		if (x<width) break;
	}
	height=y+1;

	// skip upper empty lines
	for (y=0;y<height;y++) {
		for (x=0;x<width;x++) {
			if (mask[(x+y*width)*2] || mask[(x+y*width)*2+1]) {
				break;
			}
		}
		if (x<width) break;
	}
	skiplines_ref=skiplines=y;

	/* calculer et produire le flux total */
	for (;y<height;y++) {
		// zig
		for (x=0;x<width;x++) {
			if (mask[(x+y*width)*2] || mask[(x+y*width)*2+1]) {
				// on a au moins un pixel sur l'octet à afficher!
				pix=sprite[x+y*width];
				if (!mask[(x+y*width)*2]) {
					pix&=DROITE;
					flux[iflux++]=GAUCHE;
					if (pix) flux[iflux++]=pix;
				} else if (!mask[(x+y*width)*2+1]) {
					pix&=GAUCHE;
					flux[iflux++]=DROITE;
					if (pix) flux[iflux++]=pix;
				} else {
					flux[iflux++]=pix;
				}
			}
		}
		linecpt++;
		// zag
		y++;
		if (y<height) {
			for (x=width-1;x>=0;x--) {
				if (mask[(x+y*width)*2] || mask[(x+y*width)*2+1]) {
					// on a au moins un pixel sur l'octet à afficher!
					pix=sprite[x+y*width];
					if (!mask[(x+y*width)*2]) {
						pix&=DROITE;
						flux[iflux++]=GAUCHE;
						if (pix) flux[iflux++]=pix;
					} else if (!mask[(x+y*width)*2+1]) {
						pix&=GAUCHE;
						flux[iflux++]=DROITE;
						if (pix) flux[iflux++]=pix;
					} else {
						flux[iflux++]=pix;
					}
				}
			}
		}
	}

if (debuginfo) {for (x=0;x<iflux;x++) fprintf(stderr,"#%02X,",flux[x]);fprintf(stderr,"\n");}

	mflux=iflux;
	iflux=0;
	diff_printf(output,lenoutput,";*** compiled sprite %dx%d flux=%d***\n",width,height,mflux);
	//diff_printf(output,lenoutput,"%s\n",action->label);


	// skip empty lines from top
	if (skiplines) {
		// jump lower
		int adjust=0;
		while (skiplines>=8) {
			adjust+=80;
			skiplines-=8;
		}
		if (adjust) {
			diff_printf(output,lenoutput,"ld de,%d : add hl,de ; skip block%s\n",adjust,adjust>1?"s":"");
		}
		for (y=0;y<skiplines;y++) {
			diff_printf(output,lenoutput,"rst 0 ; skip line\n");
		}
	}

	skiptrigger=0;
	linecpt=0;
	sens=1;
	x=0;
	for (y=skiplines_ref;y<height;y++) {
		ComputeDiffStats(output,lenoutput,flux,iflux,mflux,current_reg);

		// init de la position destination
		switch (sens) {
			default:
			case 1:dx=0;break;
			case -1:dx=width-1;break;
		}

		diff_printf(output,lenoutput,"/* line %03d */ ",y);

		while (1) {
			// chercher le premier pixel avec des données dans le sens de lecture
			while (!mask[(dx+y*width)*2] && !mask[(dx+y*width)*2+1]) {
				dx+=sens;
				if (dx<0 || dx>=width) {
					// on a terminé sur la ligne!
					break;
				}
			}
			if (dx<0 || dx>=width) break;

			// effectuer le déplacement vers la cible
			if (abs(dx-x)>3) {
				if (dx-x>0) diff_printf(output,lenoutput," : ld a,l : add %d : ld l,a : ld a,h : adc 0 : ld h,a",dx-x);
				else diff_printf(output,lenoutput," : ld a,l : add %d : ld l,a : ld a,h : adc #FF : ld h,a",dx-x);
				
				x=dx;
			}
			while (x<dx) {
				if ((x&3)==3) diff_printf(output,lenoutput," : inc hl : ");
				else diff_printf(output,lenoutput," : inc l : ");
				x++;
			}
			while (x>dx) {
				if (x&3) diff_printf(output,lenoutput," : dec l : ");
				else diff_printf(output,lenoutput," : dec hl : ");
				x--;
			}

//if (debuginfo) {int iir;if (y==24) {fprintf(stderr,"iflux=%d mflux=%d current_reg ",iflux,mflux);for (iir=0;iir<NBREG;iir++) fprintf(stderr,"#%02X ",current_reg[iir]);fprintf(stderr,"\n");}}
			pix=sprite[x+y*width];
			if (mask[(x+y*width)*2] && mask[(x+y*width)*2+1]) {
				// octet plein
				diff_printf(output,lenoutput," : ld (hl),%s",GetValStr(txtbuffer,current_reg,pix));iflux++;
				if (!skiptrigger) ComputeDiffStats(output,lenoutput,flux,iflux,mflux,current_reg);
			} else if (mask[(x+y*width)*2]) {
				// data à gauche
				diff_printf(output,lenoutput," : ld a,(hl) : and %s : ",GetValStr(txtbuffer,current_reg,DROITE));iflux++;
				if (!skiptrigger) ComputeDiffStats(output,lenoutput,flux,iflux,mflux,current_reg);
				if (pix&GAUCHE) {
					diff_printf(output,lenoutput,"or %s : ld (hl),a\n",GetValStr(txtbuffer,current_reg,pix&GAUCHE));iflux++;
					if (!skiptrigger) ComputeDiffStats(output,lenoutput,flux,iflux,mflux,current_reg);
				}
				else diff_printf(output,lenoutput,"ld (hl),a");
			} else if (mask[(x+y*width)*2+1]) {
				// data à droite
				diff_printf(output,lenoutput," : ld a,(hl) : and %s : ",GetValStr(txtbuffer,current_reg,GAUCHE));iflux++;
				if (!skiptrigger) ComputeDiffStats(output,lenoutput,flux,iflux,mflux,current_reg);
				if (pix&DROITE) {
					diff_printf(output,lenoutput,"or %s : ld (hl),a\n",GetValStr(txtbuffer,current_reg,pix&DROITE));iflux++;
					if (!skiptrigger) ComputeDiffStats(output,lenoutput,flux,iflux,mflux,current_reg);
				}
				else diff_printf(output,lenoutput,"ld (hl),a");
			}
			dx+=sens;
		}

		// on se passera des sauts de ligne sur la dernière
		if (y<height-1) {
			diff_printf(output,lenoutput," : rst 0\n");
		}

		diff_printf(output,lenoutput,"\n");

		// inverser le sens pour la ligne suivante
		sens=-sens;
	}
	diff_printf(output,lenoutput,"ret\n");
	diff_printf(output,lenoutput,"\n");
	diff_printf(output,lenoutput,"save './codgen/%s.bin',%s,$-%s\n",action->label,action->label,action->label);
	diff_printf(output,lenoutput,"bank\n");
	free(flux);

if (debuginfo) printf("\n");
	if (iflux!=mflux) {
		fprintf(stderr,"iflux!=mflux (%d!=%d) | [%s]\n",iflux,mflux,action->label);
		exit(-1);
	} else {
		diff_printf(output,lenoutput,";iflux/mflux %d/%d\n",iflux,mflux);
	}
	//free(sprite);
	//free(mask);
}

int FileExists(char *filename) {
	FILE *f;
	f=fopen(filename,"rb");
	if (!f) return 0;
	fclose(f);
	return 1;
}

struct s_spritesheet {
	char *label;
	unsigned char *sprite;
	unsigned char *mask;
        int count,sx,sy;
};

char **splitWithChar(char *in_str, char split_char)
{
        char **tab=NULL;
        char *match_str;
        int redo=1;
        int idx,idmax;

        match_str=in_str;

        while (redo && *in_str) {
                while (*match_str && *match_str!=split_char) match_str++;
                redo=*match_str;
                *match_str=0;
                FieldArrayAddDynamicValueConcat(&tab,&idx,&idmax,in_str);
                in_str=++match_str;
        }

        return tab;
}


void Compiler(struct s_parameter *parameter)
{
	#undef FUNC
	#define FUNC "Compiler"

	/* script */
	char scriptline[2048];
	char filename[2048];
	char label[256]; // enough
	int width,height,rwidth;
	FILE *fdef;

	/* data */
	unsigned char *data1,*data2,*col1,*col2,*ndata1,*ndata2;
	int i,j,ok=1,idx=0,x,y,cury;
	int first=-1,last,iauto=0;
	/* multi-thread */
	struct s_compilation_action *compilation_actions=NULL;
	struct s_compilation_action compilation_action={0};
	int nbcaction=0,maxcaction=0;
	struct s_compilation_thread *ct;
	int nb_cores=0;
	char *ptrlow,*ptrhigh,*descr,*preview,*descr_idx,*descr_str;

	/**/
	int lenoutput=0;
	char *output=NULL;

	// entities
        char strbuffer[1024]={0};
        FILE *f,*entities;
        char *current_category=NULL;
        struct s_spritesheet curspritesheet={0};
        struct s_spritesheet *spritesheet=NULL;
        int nbb=0,maxb=0,inbb;
        int category_idx=-2;
	int utility_categ=0;
	int road_categ=-1;
	int house_categ=-1;

	int sort_energy=-1;
	int sort_water=-1;
	int sort_recycling=-1;
	int sort_healthcare=-1;
	int sort_education=-1;
	int sort_security=-1;
	int sort_leisure=-1;

	//******************************************************************************************
	//******************************************************************************************
	//                                Entities parsing
	//******************************************************************************************
	//******************************************************************************************
	fprintf(stderr,"opening sprites.txt\n");
        f=fopen("sprites.txt","rb");
        if (!f) exit(-1);

	while (fgets(strbuffer,sizeof(strbuffer)-1,f)) {
		char **myw;
		int nbw;
		FILE *fifi;
		char filname[2048];

		myw=TxtSplitWithChar(strbuffer,' ',&nbw);
		for (idx=0;idx<nbw;idx++) {
			if (myw[idx][strlen(myw[idx])-1]==0x0D) myw[idx][strlen(myw[idx])-1]=0;
			if (myw[idx][strlen(myw[idx])-1]==0x0A) myw[idx][strlen(myw[idx])-1]=0;
			if (myw[idx][strlen(myw[idx])-1]==0x0D) myw[idx][strlen(myw[idx])-1]=0;
			if (myw[idx][strlen(myw[idx])-1]==0x0A) myw[idx][strlen(myw[idx])-1]=0;
			if (myw[idx][strlen(myw[idx])-1]==0x0D) myw[idx][strlen(myw[idx])-1]=0;
			if (myw[idx][strlen(myw[idx])-1]==0x0A) myw[idx][strlen(myw[idx])-1]=0;
		}

		memset(&curspritesheet,0,sizeof(curspritesheet));

		if (nbw<4) {
			fprintf(stderr,"skipping line with %d args\n",nbw);
			continue;
		}
		curspritesheet.label=strdup(myw[0]);
		curspritesheet.sx=atoi(myw[1]);
		curspritesheet.sy=atoi(myw[2]);
		curspritesheet.count=atoi(myw[3]);
		if (!curspritesheet.sx || !curspritesheet.sy || !curspritesheet.count) {
			fprintf(stderr,"invalid data in sprites.txt\n");
			exit(-1);
		}
		// sprite data
		sprintf(filname,"../assets/%s.bin",myw[0]);
		fifi=fopen(filname,"rb");
		if (!fifi) {
			fprintf(stderr,"file [%s] not found\n",filname);
			exit(-1);
		}
		curspritesheet.sprite=malloc(curspritesheet.sx*curspritesheet.sy*curspritesheet.count);
		curspritesheet.mask=malloc(curspritesheet.sx*curspritesheet.sy*curspritesheet.count*2);
		fread(curspritesheet.sprite,1,curspritesheet.sx*curspritesheet.sy*curspritesheet.count,fifi);
		fclose(fifi);
		// mask
		sprintf(filname,"../assets/%s.t01",myw[0]);
		fifi=fopen(filname,"rb");
		if (!fifi) {
			fprintf(stderr,"file [%s] not found\n",filname);
			exit(-1);
		}
		fread(curspritesheet.mask,1,curspritesheet.sx*curspritesheet.sy*curspritesheet.count*2,fifi);
		fclose(fifi);

		ObjectArrayAddDynamicValueConcat((void **)&spritesheet,&nbb,&maxb,&curspritesheet,sizeof(struct s_spritesheet));
		FreeWords(myw);
	}
        fclose(f);
	fprintf(stderr,"%d spritesheet created\n",nbb);



	//******************************************************************************************
	//******************************************************************************************
	//                                Sprite compilation
	//******************************************************************************************
	//******************************************************************************************

	for (inbb=0;inbb<nbb;inbb++) {
		width=spritesheet[inbb].sx;
		height=spritesheet[inbb].sy;

		for (i=0;i<spritesheet[inbb].count;i++) {
			char iLabel[256];
			int x,y,idx1,idx2;
			unsigned char vswap,p1,p2;

			ndata1=malloc(width*height);
			ndata2=malloc(width*height*2);
			memcpy(ndata1,spritesheet[inbb].sprite+width*height*i,width*height);
			memcpy(ndata2,spritesheet[inbb].mask+width*height*i*2,width*height*2);

			sprintf(iLabel,"%s_Right_%d",spritesheet[inbb].label,i);
			compilation_action.label=strdup(iLabel);
			compilation_action.sprite=ndata1;
			compilation_action.mask=ndata2;
			compilation_action.width=width;
			compilation_action.height=height;
			ObjectArrayAddDynamicValueConcat((void**)&compilation_actions,&nbcaction,&maxcaction,&compilation_action,sizeof(compilation_action));

			// Duplication for reverse swimming
			ndata1=malloc(width*height);
			ndata2=malloc(width*height*2);
			memcpy(ndata1,spritesheet[inbb].sprite+width*height*i,width*height);
			memcpy(ndata2,spritesheet[inbb].mask+width*height*i*2,width*height*2);
			for (y=0;y<height;y++) {
				idx1=y*width;
				idx2=idx1+width-1;
				while (idx1<idx2) {
					vswap=ndata1[idx1];
					ndata1[idx1]=ndata1[idx2];
					ndata1[idx2]=vswap;
					idx1++;
					idx2--;
				}
				idx1=y*width*2;
				idx2=idx1+width*2-1;
				while (idx1<idx2) {
					vswap=ndata2[idx1];
					ndata2[idx1]=ndata2[idx2];
					ndata2[idx2]=vswap;
					idx1++;
					idx2--;
				}
			}
			// reverse bits
			for (x=0;x<width*height;x++) {
				p1=ndata1[x]&0xAA;
				p2=ndata1[x]&0x55;
				p1>>=1;
				p2<<=1;
				ndata1[x]=p1|p2;
			}

			sprintf(iLabel,"%s_Left_%d",spritesheet[inbb].label,i);
			compilation_action.label=strdup(iLabel);
			compilation_action.sprite=ndata1;
			compilation_action.mask=ndata2;
			compilation_action.width=width;
			compilation_action.height=height;
			ObjectArrayAddDynamicValueConcat((void**)&compilation_actions,&nbcaction,&maxcaction,&compilation_action,sizeof(compilation_action));
		}
	}

	/* multi-thread execution */
	ct=SplitForThreads(nbcaction,compilation_actions,&nb_cores);
	ExecuteThreads(nb_cores,ct,MakeDiffThread);

	/* output assembly in legacy order */
	fdef=fopen("spriteGen.asm","wb");
	for (i=0;i<nb_cores;i++) {
		if (ct[i].output) {
			fwrite(ct[i].output,1,ct[i].lenoutput,fdef);
			fprintf(fdef,"\n");
			free(ct[i].output);
		}
		if (ct[i].nbaction) free(ct[i].action);
	}
	fclose(fdef);
	free(ct);
}

/***************************************
	semi-generic body of program
***************************************/

/*
	Usage
	display the mandatory parameters
*/
void Usage()
{
	#undef FUNC
	#define FUNC "Usage"
	
	printf("%.*s.exe v1.0 / Edouard BERGE 2020-08 Leon Marchand\n",(int)(sizeof(__FILENAME__)-3),__FILENAME__);
	printf("\n");
	printf("syntaxe is: %.*s\n",(int)(sizeof(__FILENAME__)-3),__FILENAME__);
	printf("sprites.txt must be in current path\n");
	printf("\n");
	
	exit(-1);
}

/*
	ParseOptions
	
	used to parse command line and configuration file
*/
int ParseOptions(char **argv,int argc, struct s_parameter *parameter)
{
	#undef FUNC
	#define FUNC "ParseOptions"

	int i=0;

	if (argv[i][0]=='-') {
		switch(argv[i][1])
		{
			case 'H':
			case 'h':Usage();
			default:
				Usage();				
		}
	}
	return i;
}

/*
	GetParametersFromCommandLine	
	retrieve parameters from command line and fill pointers to file names
*/
void GetParametersFromCommandLine(int argc, char **argv, struct s_parameter *parameter)
{
	#undef FUNC
	#define FUNC "GetParametersFromCommandLine"
	int i;
	
	for (i=1;i<argc;i++) {
		i+=ParseOptions(&argv[i],argc-i,parameter);
	}
}

/*
	main
	
	check parameters
	execute the main processing
*/
void main(int argc, char **argv)
{
	#undef FUNC
	#define FUNC "main"

	struct s_parameter parameter={0};

	GetParametersFromCommandLine(argc,argv,&parameter);
	Compiler(&parameter);
	exit(0);
}


